class ThorixError(Exception):
    """Base exception for all Thorix errors"""


class ThorixHttpError(ThorixError):
    """Base exception for all Thorix http errors"""


class ThorixHTTPStatusError(ThorixHttpError):
    """HTTP response had non retriable status code"""


class ThorixHTTPRetryError(ThorixHttpError):
    """HTTP Retry attempts were exhausted"""


class ThorixHTTPInvalidResponseError(ThorixHttpError):
    """HTTP response was invalid"""



from abc import ABC, abstractmethod
from typing import Any, Mapping

Json = dict[str, Any] | list[Any] | None


class AsyncTransport(ABC):
    @abstractmethod
    async def get_json(
        self, path: str, *, params: Mapping[str, Any] | None = None
    ) -> Json: ...

    @abstractmethod
    async def post_json(self, path: str, *, body: Mapping[str, Any]) -> Json: ...

    @abstractmethod
    async def aclose(self) -> None: ...


class SyncTransport(ABC):
    @abstractmethod
    def get_json(
        self, path: str, *, params: Mapping[str, Any] | None = None
    ) -> Json: ...

    @abstractmethod
    def post_json(self, path: str, *, body: Mapping[str, Any]) -> Json: ...

    @abstractmethod
    def close(self) -> None: ...



import logging
from typing import Any, Mapping

import httpx

from thorix.config.http_config import HTTPConfig
from thorix.errors import ThorixHTTPInvalidResponseError

from .retry import retry_async
from .transport import AsyncTransport, Json

logger = logging.getLogger(__name__)


class AsyncHttpTransport(AsyncTransport):
    """
    Asynchronous HTTP transport
    """

    def __init__(
        self, base_url: str, config: HTTPConfig, client: httpx.AsyncClient | None = None
    ) -> None:
        self._client = client or httpx.AsyncClient(base_url=base_url)
        self._config = config

    async def get_json(
        self, path: str, *, params: Mapping[str, Any] | None = None
    ) -> Json:
        """
        Perform a GET request with retry, returning parsed JSON
        """
        logger.debug("GET %s params=%s", path, params)

        async def _do() -> Json:
            r = await self._client.get(path, params=params)
            r.raise_for_status()
            try:
                return r.json()
            except ValueError as exc:
                raise ThorixHTTPInvalidResponseError(
                    "Invalid JSON response from Thor node"
                ) from exc

        return await retry_async(_do, self._config)

    async def post_json(self, path: str, *, body: Mapping[str, Any]) -> Json:
        """
        Perform a POST request with retry, returning parsed JSON
        """
        logger.debug("POST %s payload=%s", path, body)

        async def _do() -> Json:
            r = await self._client.post(path, json=body)
            r.raise_for_status()
            try:
                return r.json()
            except ValueError as exc:
                raise ThorixHTTPInvalidResponseError(
                    "Invalid JSON response from Thor node"
                ) from exc

        return await retry_async(_do, self._config)

    async def aclose(self) -> None:
        """
        Closes the httpx client
        """
        await self._client.aclose()



import asyncio
import logging
import random
import time
from typing import Awaitable, Callable

import httpx

from thorix.config.http_config import HTTPConfig
from thorix.errors import ThorixHTTPRetryError, ThorixHTTPStatusError
from thorix.http.transport import Json

logger = logging.getLogger(__name__)


def _compute_delay(attempt: int, config: HTTPConfig) -> float:
    """
    Compute exponential backoff delay with optional cap and jitter.
    """
    delay = config.retry_base_delay * (2**attempt)
    if config.retry_max_delay is not None:
        delay = min(delay, config.retry_max_delay)
    return delay + random.random() * config.retry_delay_jitter


async def retry_async(fn: Callable[[], Awaitable[Json]], config: HTTPConfig) -> Json:
    """
    Async retry, will retry if it received:
        - RequestError
        - HttpStatusCodeError and status code is >= 500
    It uses an exponential back-off strategy
    Note total attempts is max_retries + 1 (the initial attempt)
    """
    for attempt in range(config.max_retries + 1):
        try:
            return await fn()
        except httpx.RequestError as exc:
            if attempt == config.max_retries:
                logger.error(
                    "Retry attempts exhausted after request error",
                    exc_info=exc,
                )
                raise ThorixHTTPRetryError(
                    "Retry attempts exhausted after request error"
                ) from exc
            delay = _compute_delay(attempt, config)
            logger.warning(
                "HTTP request failed (%s). Retrying attempt %d/%d in %.2fs",
                exc.__class__.__name__,
                attempt + 1,
                config.max_retries,
                delay,
            )
            await asyncio.sleep(delay)
        except httpx.HTTPStatusError as exc:
            status = exc.response.status_code
            if status < 500:
                logger.error(
                    "HTTP request failed with status %s",
                    status,
                    exc_info=exc,
                )
                raise ThorixHTTPStatusError(
                    f"HTTP request failed with status {status}"
                ) from exc
            if attempt == config.max_retries:
                logger.error(
                    "Retry attempts exhausted after server error %s",
                    status,
                    exc_info=exc,
                )
                raise ThorixHTTPRetryError(
                    f"Retry attempts exhausted after server error {status}"
                ) from exc
            delay = _compute_delay(attempt, config)
            logger.warning(
                "Server error %s. Retrying attempt %d/%d in %.2fs",
                status,
                attempt + 1,
                config.max_retries - 1,
                delay,
            )
            await asyncio.sleep(delay)
    raise ThorixHTTPRetryError("Retry attempts exhausted")


def retry_sync(fn: Callable[[], Json], config: HTTPConfig) -> Json:
    """
    Sync retry, will retry if it receives:
        - RequestError
        - HttpStatusCodeError and status code >= 500
    It uses an exponential back-off strategy
    Note total attempts is max_retries + 1 (the initial attempt)
    """
    for attempt in range(config.max_retries + 1):
        try:
            return fn()
        except httpx.RequestError as exc:
            if attempt == config.max_retries:
                logger.error(
                    "Retry attempts exhausted after request error",
                    exc_info=exc,
                )
                raise ThorixHTTPRetryError(
                    "Retry attempts exhausted after request error"
                ) from exc
            delay = _compute_delay(attempt, config)
            logger.warning(
                "HTTP request failed (%s). Retrying attempt %d/%d in %.2fs",
                exc.__class__.__name__,
                attempt + 1,
                config.max_retries,
                delay,
            )
            time.sleep(delay)
        except httpx.HTTPStatusError as exc:
            status = exc.response.status_code
            if status < 500:
                logger.error(
                    "HTTP request failed with status %s",
                    status,
                    exc_info=exc,
                )
                raise ThorixHTTPStatusError(
                    f"HTTP request failed with status {status}"
                ) from exc
            if attempt == config.max_retries:
                logger.error(
                    "Retry attempts exhausted after server error %s",
                    status,
                    exc_info=exc,
                )
                raise ThorixHTTPRetryError(
                    f"Retry attempts exhausted after server error {status}"
                ) from exc
            delay = _compute_delay(attempt, config)
            logger.warning(
                "Server error %s. Retrying attempt %d/%d in %.2fs",
                status,
                attempt + 1,
                config.max_retries - 1,
                delay,
            )
            time.sleep(delay)
    raise ThorixHTTPRetryError("Retry attempts exhausted")



from dataclasses import dataclass


@dataclass(frozen=True)
class HTTPConfig:
    max_retries: int = 3
    retry_base_delay: float = 0.5
    retry_max_delay: float | None = None
    retry_delay_jitter: float = 0.1
